/*** simple LOKI (or LOKI alike) detection routines.
 *** (C) 1999/2000 by Stealth. GPL.
 ***/
#include "imon.h"

struct virtual_connections {
   	int replys, requests;
        unsigned int src, dst;
        struct virtual_connections *next;
};

struct virtual_connections *first = NULL;

void alarm_handler(int);

void
passive_loki(char *intf, int timeout)
{
   	int fd = 0;
        char buf[10000];
        struct icmphdr *icmp;
        struct iphdr *ip;
	unsigned int t = 0;

	printf("Try to detect active LOKI sessions. You gave me %d seconds.\n"
	       "Please be patient.\n\n", timeout);

        /* set handler to interrupt us later */
   	signal(SIGALRM, alarm_handler);

   	/* give us a socket */
        if ((fd = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
           	perror("socket");
                exit(errno);
        }
        setpromisc(fd, intf);     
        promisc = fd;
        alarm(timeout);        
        while (1) {
           	bzero(buf, 10000);
           	if (read(fd, buf, 9999) < 0 && errno != EINTR) {
                   	perror("read");
                        exit(errno);
                }
                /* point to IP and ICMP headers in received packet */
                ip = (struct iphdr*)(buf + sizeof(struct ethhdr));
                icmp = (struct icmphdr*)(buf + sizeof(struct ethhdr) + (ip->ihl<<2));
                
                /* only handle ICMP packets */
                if (ip->protocol != ICMPv4)
                   	continue;
                /* from type 0 and type 8 */
                if (icmp->type != ICMP_ECHOREPLY && icmp->type != ICMP_ECHO)
                   	continue;
		/* swap SRC/DST if REPLY, or we count the connections doubled */
		if (icmp->type == ICMP_ECHOREPLY) {
			t = ip->saddr;
			ip->saddr = ip->daddr;
			ip->daddr = t;
		}
                add_connection(ip->saddr,ip->daddr, icmp->type);
        }                
}

int
add_connection(unsigned int src, unsigned int dst, int what)
{
   	struct virtual_connections *l;
        
        l = first;
        
        /* look if connection already exists */
        while (l) {
           	/* if so, count type */
           	if (l->src == src && l->dst == dst) {
                   	if (what == ICMP_ECHO) {
#ifdef DEBUG
				printf("requests++\n");
#endif
                           	l->requests++;
			} else {
#ifdef DEBUG
				printf("replys++\n");
#endif
                           	l->replys++;
			}
                   	return 0;
                }
                l = l->next;
        }
        /* Ok, here we are. Thus the connection doesn't exist */
        if ((l = (struct virtual_connections*)malloc(sizeof(struct virtual_connections))) == NULL) {
                perror("malloc");
                exit(errno);
        }
	memset(l, 0, sizeof(struct virtual_connections));
        if (what == ICMP_ECHO)
           	l->requests++;
        else
           	l->replys++;
        /* linked list: next entry */
        l->src = src;
        l->dst = dst;
        l->next = first;
#ifdef DEBUG
	printf("%d %d\n", src, dst);
#endif
	
        first = l;
        return 0;
}

/* called from passive_loki() upon exit.
 * passive_loki() should never 'return' .
 */
void
alarm_handler(int sig)
{
	char *s1 = NULL, *s2 = NULL;
	int bad = 0;
	
   	while (first) {
#ifdef DEBUG
           	printf("%d %d\n", first->replys, first->requests);
#endif
		if ((first->replys / first->requests) > MAXREPLYS) {
			s1 = resolv(first->dst, 1);
			s2 = resolv(first->src, 1);
			printf("\a\n*******\n");	
			printf("Alert! Possibly LOKI session detected from %s to %s!\n", s2, s1);
			printf("(attacked host is %s)\n", s1);
			printf("*******\n\n");
			bad = 1;
		}	
                first = first->next;
        }
	if (!bad) 
		printf("No active sessions detected. Ok.\n\n");
	free(s1);
	free(s2);
        droppromisc(promisc);
        exit(0);
}

/* Scan a host active via special ICMP packets.
 * Just like a virusscanner ;-)
 * Use device 'intf' for it with 'to' as the host and
 * 'from' for the (maybe spoofed) source adress.
 */
int 
active_loki(char *intf, char *to, char *from)
{
   	struct icmphdr *icmp = NULL;
        struct iphdr *ip = NULL;
        int rfd = 0, while_count = 1, datalen = 0;
        int one = 1, count = 0;
        struct sockaddr_in to_addr, from_addr;
        struct hostent *he;
        char buf[10000] = {0};
	void *send_buf = NULL;
        struct timeval tv;
        char *payload =  NULL;
	fd_set r_set;
        
	/* resolve hostnames to in_addr */
        if ((he = gethostbyname(to)) == NULL) {
		perror("gethostbyname");
		exit(errno);
	} else 
           	memcpy(&(to_addr.sin_addr), he->h_addr, he->h_length);
		
        if ((he = gethostbyname(from)) == NULL) {
		perror("gethostbyname");
           	exit(errno);
	} else 
           	memcpy(&(from_addr.sin_addr), he->h_addr, he->h_length);

	/* get some memory */
	if ((send_buf = malloc(10000)) == NULL) {
		perror("malloc");
		exit(errno);
	}
        memset(send_buf, 0, 10000);
	
	if ((rfd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
           	perror("socket");
                exit(errno);
        }
        setpromisc(rfd, intf);
        promisc = rfd;
        
        /* let us write IP headers 
        if (setsockopt(rfd, IPPROTO_IP, IP_HDRINCL, &one, sizeof(one)) < 0) {
           	perror("setsockopt");
                exit(errno);
        }*/
        ip = (struct iphdr*)send_buf;
        icmp = (struct icmphdr*)send_buf;
	
	datalen = sizeof(struct icmphdr) + 5;
        
	/* build headers for send */
        icmp->type = ICMP_ECHO;		/* C_P_TYPE */
        icmp->code = 0;			
        icmp->un.echo.id = 11;
        icmp->un.echo.sequence = 0xf001;		/* L_TAG */
	payload = send_buf+sizeof(struct icmphdr); /* + sizeof(struct iphdr); */
	
	payload[0] = 0xb1;		/* L_REQ */
	
	payload[1] = 0x15;		/* some data, "ls" */
	payload[2] = 0x79;		/* propably obsolete ...*/
	payload[3] = 0xa;
	payload[4] = 0;
        
	icmp->checksum = in_cksum((u_short*)icmp, datalen);
	

	to_addr.sin_family = AF_INET;
	to_addr.sin_port = 0;
	
	/* send _one_ packet */
        if (sendto(rfd, send_buf, datalen, 0, (struct sockaddr*)&to_addr, sizeof(to_addr)) < 0) {
		perror("sendto");
		free(send_buf);
		exit(errno);
    	}
        
        /* wait for answers to receive */
        while (count < 2) {
           	FD_ZERO(&r_set);
                FD_SET(rfd, &r_set);
                
                /* wait 3 seconds for the first reply, propably from
                 * the kernel
                 */
                if (while_count == 1)
                   	tv.tv_sec =  3;
                /* but only 1 sec for the possible 2nd packet, since it
                 * is send directly after the first packet, it has 4 seconds time
                 * at all - should be enuff
                 */
                else
                   	tv.tv_sec = 1;
                tv.tv_usec = 0;
                
                /* blocking */
                select(rfd + 1, &r_set, NULL, NULL, &tv);
                
                /* if no packets arrive within the time, leave loop */
                if (!FD_ISSET(rfd, &r_set))
                   	break;
                /* non-blocking read due to select() */
                if (read(rfd, buf, 9999) < 0)
			break;
		ip = (struct iphdr*)buf;
		                
                /* is it the right packet ? */
		if (ip->daddr == from_addr.sin_addr.s_addr) {
            		icmp = (struct icmphdr*)(buf +  (ip->ihl<<2));
#ifdef DEBUG
			printf("%d\n", icmp->type);
#endif
			if (icmp->type == ICMP_ECHOREPLY)
				count++;
		}
#ifdef DEBUG
		printf("while_count++:\n");
#endif
                while_count++;
        } /* while */
        droppromisc(rfd);
	if (count > 1)  {
		printf("\aHost '%s' seems to have a ICMP backdoor!\n", to);
	} else if (count == 1) {
		printf("Host '%s' seems to be OK.\t\t\t\t\t\r", to); 
		fflush(stdout);
	} else {
           	printf("Host '%s' seems to be dead (or firewalled).\t\t\t\t\t\r", to);
		fflush(stdout);
	}
	printf("\n\n");
	free(send_buf);
        return 0;
}

        
        

